# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import hashlib
import logging
from time import sleep
from typing import List
from typing import Optional
from typing import Sequence
from typing import TypedDict

from kubernetes.client import AuthenticationV1TokenRequest
from kubernetes.client import V1Job
from kubernetes.client import V1ObjectMeta
from kubernetes.client import V1Pod
from kubernetes.client import V1PolicyRule
from kubernetes.client import V1Role
from kubernetes.client import V1RoleBinding
from kubernetes.client import V1RoleRef
from kubernetes.client import V1ServiceAccount
from kubernetes.client import V1Subject
from kubernetes.client import V1TokenRequestSpec
from kubernetes.client.exceptions import ApiException

from paasta_tools.adhoc_tools import load_adhoc_job_config
from paasta_tools.eks_tools import EksDeploymentConfig
from paasta_tools.eks_tools import load_eks_service_config
from paasta_tools.kubernetes.application.controller_wrappers import (
    get_application_wrapper,
)
from paasta_tools.kubernetes_tools import get_all_service_accounts
from paasta_tools.kubernetes_tools import JOB_TYPE_LABEL_NAME
from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.kubernetes_tools import limit_size_with_hash
from paasta_tools.kubernetes_tools import paasta_prefixed
from paasta_tools.utils import load_system_paasta_config


logger = logging.getLogger(__name__)
REMOTE_RUN_JOB_LABEL = "remote-run"
POD_OWNER_LABEL = paasta_prefixed("pod_owner")
TOOLBOX_MOCK_SERVICE = "prod-toolbox"
DEFAULT_MAX_DURATION_LIMIT = 8 * 60 * 60  # 8 hours


class RemoteRunError(Exception):
    pass


class RemoteRunOutcome(TypedDict, total=False):
    status: int
    message: str
    job_name: str
    pod_name: str
    pod_address: str
    namespace: str


def _format_remote_run_job_name(
    job: V1Job,
    user: str,
) -> str:
    """Format name for remote run job

    :param V1Job job: job definition
    :param str user: the user requesting the remote-run
    :return: job name
    """
    return limit_size_with_hash(f"remote-run-{user}-{job.metadata.name}")


def remote_run_start(
    service: str,
    instance: str,
    cluster: str,
    user: str,
    interactive: bool,
    recreate: bool,
    max_duration: int,
    is_toolbox: bool,
) -> RemoteRunOutcome:
    """Trigger remote-run job

    :param str service: service name
    :param str instance: service instance
    :param str cluster: paasta cluster
    :param str user: the user requesting the remote-run sandbox
    :param bool interactive: whether it is expected to access the remote-run job interactively
    :param bool recreate: whether to recreate remote-run job if existing
    :param int max_duration: maximum allowed duration for the remote-ruh job
    :param bool is_toolbox: requested job is for a toolbox container
    :return: outcome of the operation, and resulting Kubernetes pod information
    """
    kube_client = KubeClient()

    # Load the service deployment settings
    deployment_config = (
        generate_toolbox_deployment(service, cluster, user)
        if is_toolbox
        else load_eks_service_config(service, instance, cluster)
    )

    # Set to interactive mode
    if interactive and not is_toolbox:
        deployment_config.config_dict["cmd"] = f"sleep {max_duration}"

    # Create the app with a new name
    formatted_job = deployment_config.format_kubernetes_job(
        job_label=REMOTE_RUN_JOB_LABEL,
        deadline_seconds=max_duration,
        keep_routable_ip=is_toolbox,
    )
    job_name = _format_remote_run_job_name(formatted_job, user)
    formatted_job.metadata.name = job_name
    app_wrapper = get_application_wrapper(formatted_job)
    app_wrapper.soa_config = deployment_config

    # Launch pod
    logger.info(f"Starting {job_name}")
    try:
        app_wrapper.create(kube_client)
    except ApiException as e:
        if e.status != 409:
            raise
        if recreate:
            remote_run_stop(
                service=service,
                instance=instance,
                cluster=cluster,
                user=user,
                is_toolbox=is_toolbox,
            )
            return remote_run_start(
                service=service,
                instance=instance,
                cluster=cluster,
                user=user,
                interactive=interactive,
                recreate=False,
                max_duration=max_duration,
                is_toolbox=is_toolbox,
            )

    return {
        "status": 200,
        "message": "Remote run sandbox started",
        "job_name": job_name,
    }


def remote_run_ready(
    service: str,
    instance: str,
    cluster: str,
    job_name: str,
    user: str,
    is_toolbox: bool,
) -> RemoteRunOutcome:
    """Check if remote-run pod is ready

    :param str service: service name
    :param str instance: service instance
    :param str cluster: paasta cluster
    :param str job_name: name of the remote-run job to check
    :param bool is_toolbox: requested job is for a toolbox container
    :return: job status, with pod info
    """
    kube_client = KubeClient()

    # Load the service deployment settings
    deployment_config = (
        generate_toolbox_deployment(service, cluster, user)
        if is_toolbox
        else load_eks_service_config(service, instance, cluster)
    )
    namespace = deployment_config.get_namespace()

    pod = find_job_pod(kube_client, namespace, job_name)
    if not pod:
        return {"status": 404, "message": "No pod found"}
    if pod.status.phase == "Running":
        result: RemoteRunOutcome = {
            "status": 200,
            "message": "Pod ready",
            "pod_name": pod.metadata.name,
            "namespace": namespace,
        }
        if is_toolbox:
            result["pod_address"] = pod.status.pod_ip
        return result
    return {
        "status": 204,
        "message": "Pod not ready",
    }


def remote_run_stop(
    service: str,
    instance: str,
    cluster: str,
    user: str,
    is_toolbox: bool,
) -> RemoteRunOutcome:
    """Stop remote-run job

    :param str service: service name
    :param str instance: service instance
    :param str cluster: paasta cluster
    :param str user: the user requesting the remote-run sandbox
    :param bool is_toolbox: requested job is for a toolbox container
    :return: outcome of the operation
    """
    kube_client = KubeClient()

    # Load the service deployment settings
    deployment_config = (
        generate_toolbox_deployment(service, cluster, user)
        if is_toolbox
        else load_eks_service_config(service, instance, cluster)
    )

    # Rebuild the job metadata
    formatted_job = deployment_config.format_kubernetes_job(
        job_label=REMOTE_RUN_JOB_LABEL
    )
    job_name = _format_remote_run_job_name(formatted_job, user)
    formatted_job.metadata.name = job_name

    # Stop the job
    logger.info(f"Stopping {job_name}")
    app_wrapper = get_application_wrapper(formatted_job)
    app_wrapper.soa_config = deployment_config
    app_wrapper.deep_delete(kube_client)

    return {"status": 200, "message": "Job successfully removed"}


def remote_run_token(
    service: str,
    instance: str,
    cluster: str,
    user: str,
) -> str:
    """Creates a short lived token for execing into a pod

    :param str service: service name
    :param str instance: service instance
    :param str cluster: paasta cluster
    :param str user: the user requesting the remote-run sandbox
    """
    kube_client = KubeClient()

    # Load the service deployment settings
    deployment_config = load_eks_service_config(service, instance, cluster)
    namespace = deployment_config.get_namespace()

    # Rebuild the job metadata
    formatted_job = deployment_config.format_kubernetes_job(
        job_label=REMOTE_RUN_JOB_LABEL
    )
    job_name = _format_remote_run_job_name(formatted_job, user)

    # Find pod and create exec token for it
    pod = find_job_pod(kube_client, namespace, job_name)
    if not pod:
        raise RemoteRunError(f"Pod for {job_name} not found")
    pod_name = pod.metadata.name
    logger.info(f"Generating temporary service account token for {pod_name}")
    service_account = create_remote_run_service_account(
        kube_client, namespace, pod_name, user
    )
    role = create_pod_scoped_role(kube_client, namespace, pod_name, user)
    bind_role_to_service_account(kube_client, namespace, service_account, role, user)
    return create_temp_exec_token(kube_client, namespace, service_account)


def generate_toolbox_deployment(
    service: str, cluster: str, user: str
) -> EksDeploymentConfig:
    """Creates virtual EKS deployment for toolbox containers starting from adhoc configuration

    :param str service: toolbox name
    :param str cluster: target deployment cluster
    :param str user: user requesting the toolbox
    :return: deployment configuration
    """
    if not user.isalnum():
        raise RemoteRunError(
            f"Provided username contains non-alphanumeric characters: {user}"
        )
    # NOTE: API authorization is enforced by service, and we want different rules
    # for each toolbox, so clients send a combined service-instance string, and then
    # we split it here to load the correct instance settings.
    adhoc_instance = service[len(TOOLBOX_MOCK_SERVICE) + 1 :]
    adhoc_deployment = load_adhoc_job_config(
        TOOLBOX_MOCK_SERVICE,
        adhoc_instance,
        cluster,
        load_deployments=False,
    )
    # NOTE: we're explicitly dynamically mounting a single user's public keys
    # as we want these pods to only be usable by said user.
    adhoc_deployment.config_dict.setdefault("extra_volumes", []).append(
        {
            "containerPath": f"/etc/authorized_keys.d/{user}.pub",
            "hostPath": f"/etc/authorized_keys.d/{user}.pub",
            "mode": "RO",
        },
    )
    adhoc_deployment.config_dict.setdefault("env", {})["SANDBOX_USER"] = user
    adhoc_deployment.config_dict["routable_ip"] = True
    return EksDeploymentConfig(
        service=service,
        cluster=cluster,
        instance="main",
        config_dict=adhoc_deployment.config_dict,
        branch_dict=adhoc_deployment.branch_dict,
    )


def find_job_pod(
    kube_client: KubeClient,
    namespace: str,
    job_name: str,
    job_label: str = REMOTE_RUN_JOB_LABEL,
    retries: int = 3,
) -> Optional[V1Pod]:
    """Locate pod for remote-run job

    :param KubeClient kube_client: Kubernetes client
    :param str namespace: the pod namespace
    :param str job_name: remote-run job name
    :param int retries: maximum number of attemps
    :return: pod object if found
    """
    selectors = (
        f"{paasta_prefixed(JOB_TYPE_LABEL_NAME)}={job_label}",
        f"job-name={job_name}",
    )
    for _ in range(retries):
        pod_list = kube_client.core.list_namespaced_pod(
            namespace,
            label_selector=",".join(selectors),
        )
        if pod_list.items:
            return pod_list.items[0]
        sleep(0.5)
    return None


def create_temp_exec_token(
    kube_client: KubeClient,
    namespace: str,
    service_account: str,
) -> str:
    """Create a short lived token for service account

    :param KubeClient kube_client: Kubernetes client
    :param str namespace: service account namespace
    :param str service_account: service account name
    :return: token value
    """
    token_spec = V1TokenRequestSpec(
        expiration_seconds=600,  # minimum allowed by k8s
        audiences=[],
    )
    request = AuthenticationV1TokenRequest(spec=token_spec)
    response = kube_client.core.create_namespaced_service_account_token(
        service_account, namespace, request
    )
    return response.status.token


def get_remote_run_service_accounts(
    kube_client: KubeClient, namespace: str, user: str = ""
) -> Sequence[V1ServiceAccount]:
    """List all temporary service account related to remote-run

    :param KubeClient kube_client: Kubernetes client
    :param str namespace: pod namespace
    :param str user: optionally filter by owning user
    :return: list of service accounts
    """
    return get_all_service_accounts(
        kube_client,
        namespace=namespace,
        label_selector=(f"{POD_OWNER_LABEL}={user}" if user else POD_OWNER_LABEL),
    )


def create_remote_run_service_account(
    kube_client: KubeClient,
    namespace: str,
    pod_name: str,
    user: str,
) -> str:
    """Create service account to exec into remote-run pod

    :param KubeClient kube_client: Kubernetes client
    :param str namespace: pod namespace
    :param str pod_name: pod name
    :param str user: user requiring credentials
    """
    pod_name_hash = hashlib.sha1(pod_name.encode("utf-8")).hexdigest()[:12]
    service_account_name = limit_size_with_hash(f"remote-run-{user}-{pod_name_hash}")
    service_accounts = get_remote_run_service_accounts(kube_client, namespace, user)
    if any(item.metadata.name == service_account_name for item in service_accounts):
        return service_account_name
    service_account = V1ServiceAccount(
        metadata=V1ObjectMeta(
            name=service_account_name,
            namespace=namespace,
            labels={POD_OWNER_LABEL: user},
        )
    )
    kube_client.core.create_namespaced_service_account(
        namespace=namespace, body=service_account
    )
    return service_account_name


def create_pod_scoped_role(
    kube_client: KubeClient,
    namespace: str,
    pod_name: str,
    user: str,
) -> str:
    """Create role with execution access to specific pod

    :param KubeClient kube_client: Kubernetes client
    :param str namespace: pod namespace
    :param str pod_name: pod name
    :param str user: user requiring the role
    :return: name of the role
    """
    pod_name_hash = hashlib.sha1(pod_name.encode("utf-8")).hexdigest()[:12]
    role_name = f"remote-run-role-{pod_name_hash}"
    policy = V1PolicyRule(
        verbs=["create", "get"],
        resources=["pods", "pods/exec"],
        resource_names=[pod_name],
        api_groups=[""],
    )
    role = V1Role(
        rules=[policy],
        metadata=V1ObjectMeta(
            name=role_name,
            labels={POD_OWNER_LABEL: user},
        ),
    )
    try:
        kube_client.rbac.create_namespaced_role(namespace=namespace, body=role)
    except ApiException as e:
        if e.status != 409:
            raise
    return role_name


def bind_role_to_service_account(
    kube_client: KubeClient,
    namespace: str,
    service_account: str,
    role: str,
    user: str,
) -> None:
    """Bind service account to role

    :param KubeClient kube_client: Kubernetes client
    :param str namespace: service account namespace
    :param str service_account: service account name
    :param str role: role name
    :param str user: user requiring the role
    """
    role_binding = V1RoleBinding(
        metadata=V1ObjectMeta(
            name=limit_size_with_hash(f"remote-run-binding-{role}"),
            namespace=namespace,
            labels={POD_OWNER_LABEL: user},
        ),
        role_ref=V1RoleRef(
            api_group="rbac.authorization.k8s.io",
            kind="Role",
            name=role,
        ),
        subjects=[
            V1Subject(
                kind="ServiceAccount",
                name=service_account,
            ),
        ],
    )
    try:
        kube_client.rbac.create_namespaced_role_binding(
            namespace=namespace,
            body=role_binding,
        )
    except ApiException as e:
        if e.status != 409:
            raise


def get_remote_run_roles(kube_client: KubeClient, namespace: str) -> List[V1Role]:
    """List all temporary roles related to remote-run

    :param KubeClient kube_client: Kubernetes client
    :param str namespace: role namespace
    :return: list of roles
    """
    return kube_client.rbac.list_namespaced_role(
        namespace,
        label_selector=POD_OWNER_LABEL,
    ).items


def get_remote_run_role_bindings(
    kube_client: KubeClient, namespace: str
) -> List[V1RoleBinding]:
    """List all temporary role bindings related to remote-run

    :param KubeClient kube_client: Kubernetes client
    :param str namespace: role namespace
    :return: list of roles
    """
    return kube_client.rbac.list_namespaced_role_binding(
        namespace,
        label_selector=POD_OWNER_LABEL,
    ).items


def get_remote_run_jobs(kube_client: KubeClient, namespace: str) -> List[V1Job]:
    """List all remote-run jobs

    :param KubeClient kube_client: Kubernetes client
    :param str namespace: job namespace
    """
    return kube_client.batches.list_namespaced_job(
        namespace,
        label_selector=f"{paasta_prefixed(JOB_TYPE_LABEL_NAME)}={REMOTE_RUN_JOB_LABEL}",
    ).items


def get_max_job_duration_limit() -> int:
    """Get maximum configured duration for a remote run job

    :return: max duration in seconds
    """
    system_config = load_system_paasta_config()
    return system_config.get_remote_run_duration_limit(DEFAULT_MAX_DURATION_LIMIT)



import argparse
import logging

from paasta_tools.config_utils import AutoConfigUpdater
from paasta_tools.utils import DEFAULT_SOA_CONFIGS_GIT_URL
from paasta_tools.utils import format_git_url
from paasta_tools.utils import load_system_paasta_config

log = logging.getLogger(__name__)


def get_default_git_remote() -> str:
    system_paasta_config = load_system_paasta_config()
    repo_config = system_paasta_config.get_git_repo_config("yelpsoa-configs")
    default_git_remote = format_git_url(
        system_paasta_config.get_git_config()["git_user"],
        repo_config.get("git_server", DEFAULT_SOA_CONFIGS_GIT_URL),
        repo_config["repo_name"],
    )
    return default_git_remote


def parse_args():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument(
        "--git-remote",
        help="Master git repo for soaconfigs",
        default=get_default_git_remote(),
        dest="git_remote",
    )
    parser.add_argument(
        "--branch",
        help="Branch name to push to",
        required=True,
        dest="branch",
    )
    parser.add_argument(
        "--local-dir",
        help="Act on configs in the local directory rather than cloning the git_remote",
        required=False,
        default=None,
        dest="local_dir",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="Logging verbosity",
        action="store_true",
        dest="verbose",
    )
    parser.add_argument(
        "--source-id",
        help="String to attribute the changes in the commit message.",
        required=False,
        default=None,
        dest="source_id",
    )
    parser.add_argument(
        "--service",
        help="Service to modify",
        required=True,
        dest="service",
    )
    parser.add_argument(
        "--shard-name",
        help="Shard name to remove",
        required=True,
        dest="shard_name",
    )
    parser.add_argument(
        "--superregion",
        nargs="+",
        help="Superregion to remove shard from",
        required=True,
    )
    return parser.parse_args()


SUPPERREGION_DEPLOY_MAPPINGS = {
    "norcal-devc": "dev",
    "norcal-stagef": "stage",
    "norcal-stageg": "stage",
    "nova-prod": "prod",
    "pnw-prod": "prod",
}


def main(args):
    updater = AutoConfigUpdater(
        config_source=args.source_id,
        git_remote=args.git_remote,
        branch=args.branch,
        working_dir=args.local_dir or "/nail/tmp",
        do_clone=args.local_dir is None,
    )
    with updater:
        deploy_file = updater.get_existing_configs(args.service, "deploy")
        smartstack_file = updater.get_existing_configs(args.service, "smartstack")
        delete_all = set(args.superregion) == set(SUPPERREGION_DEPLOY_MAPPINGS.keys())

        for superregion in args.superregion:
            # Determine / load configuration (PAASTA-18216)
            eks_config = updater.get_existing_configs(
                args.service, f"eks-{superregion}"
            )
            kube_config = updater.get_existing_configs(
                args.service, f"kubernetes-{superregion}"
            )

            if eks_config:
                config_file = eks_config
                config_file_suffix = "eks"
            elif kube_config:
                config_file = kube_config
                config_file_suffix = "kubernetes"
            else:
                log.error(f"No EKS or Kubernetes config found for {args.service}")
                continue

            # Remove shard from deploy pipeline config
            targeted_step = (
                f"{SUPPERREGION_DEPLOY_MAPPINGS[superregion]}.{args.shard_name}"
            )
            deploy_file["pipeline"] = [
                i for i in deploy_file["pipeline"] if i["step"] != targeted_step
            ]
            log.info(f"{targeted_step} removed from deploy config")
            updater.write_configs(args.service, "deploy", deploy_file)

            # Remove shard from corresponding config
            del config_file[args.shard_name]
            updater.write_configs(
                args.service, f"{config_file_suffix}-{superregion}", config_file
            )
            log.info(
                f"{args.shard_name} removed from {config_file_suffix}-{superregion}"
            )

        # If we are removing the shard from all regions, remove it from smartstack too.
        if delete_all:
            del smartstack_file[args.shard_name]
            updater.write_configs(args.service, "smartstack", smartstack_file)
            log.info(f"{args.shard_name} removed from smartstack")
        else:
            log.info(f"not removing {args.shard_name} from smartstack")

        updater.commit_to_remote()


if __name__ == "__main__":
    args = parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARNING)

    main(args)



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

import paasta_tools.paastaapi.models as paastamodels
from paasta_tools.api import client
from paasta_tools.cli.utils import figure_out_service_name
from paasta_tools.cli.utils import get_instance_configs_for_service
from paasta_tools.cli.utils import get_paasta_oapi_api_clustername
from paasta_tools.cli.utils import lazy_choices_completer
from paasta_tools.cli.utils import list_instances
from paasta_tools.eks_tools import EksDeploymentConfig
from paasta_tools.utils import _log_audit
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import list_clusters
from paasta_tools.utils import list_services
from paasta_tools.utils import PaastaColors


log = logging.getLogger(__name__)


def add_subparser(subparsers):
    autoscale_parser = subparsers.add_parser(
        "autoscale",
        help="Manually scale a service up and down manually, bypassing the normal autoscaler",
    )

    autoscale_parser.add_argument(
        "-s",
        "--service",
        help="Service that you want to autoscale. Like 'example_service'.",
    ).completer = lazy_choices_completer(list_services)
    autoscale_parser.add_argument(
        "-i",
        "--instance",
        help="Instance of the service that you want to autoscale. Like 'main' or 'canary'.",
        required=True,
    ).completer = lazy_choices_completer(list_instances)
    autoscale_parser.add_argument(
        "-c",
        "--cluster",
        help="The PaaSTA cluster that has the service instance you want to autoscale. Like 'pnw-prod'.",
        required=True,
    ).completer = lazy_choices_completer(list_clusters)
    autoscale_parser.add_argument(
        "--set", help="Set the number to scale to. Must be an Int.", type=int
    )
    autoscale_parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    autoscale_parser.set_defaults(command=paasta_autoscale)


def paasta_autoscale(args):
    log.setLevel(logging.DEBUG)
    service = figure_out_service_name(args)
    instance_config = next(
        get_instance_configs_for_service(
            service=service,
            soa_dir=args.soa_dir,
            clusters=[args.cluster],
            instances=[args.instance],
        ),
        None,
    )
    if not instance_config:
        print(
            "Could not find config files for this service instance in soaconfigs. Maybe you mispelled an argument?"
        )
        return 1

    api = client.get_paasta_oapi_client(
        cluster=get_paasta_oapi_api_clustername(
            cluster=args.cluster,
            is_eks=(instance_config.__class__ == EksDeploymentConfig),
        ),
        http_res=True,
    )
    if not api:
        print("Could not connect to paasta api. Maybe you misspelled the cluster?")
        return 1

    try:
        if args.set is None:
            log.debug("Getting the current autoscaler count...")
            res, status, _ = api.autoscaler.get_autoscaler_count(
                service=service, instance=args.instance, _return_http_data_only=False
            )
        else:
            log.debug(f"Setting desired instances to {args.set}.")
            msg = paastamodels.AutoscalerCountMsg(desired_instances=int(args.set))
            res, status, _ = api.autoscaler.update_autoscaler_count(
                service=service,
                instance=args.instance,
                autoscaler_count_msg=msg,
                _return_http_data_only=False,
            )

            _log_audit(
                action="manual-scale",
                action_details=str(msg),
                service=service,
                instance=args.instance,
                cluster=args.cluster,
            )
    except api.api_error as exc:
        status = exc.status

    if not 200 <= status <= 299:
        print(
            PaastaColors.red(
                f"ERROR: '{args.instance}' is not configured to autoscale, "
                f"so paasta autoscale could not scale it up on demand. "
                f"If you want to be able to boost this service, please configure autoscaling for the service "
                f"in its config file by setting min and max instances. Example: \n"
                f"{args.instance}:\n"
                f"     min_instances: 5\n"
                f"     max_instances: 50"
            )
        )
        return 0

    log.debug(f"Res: {res} Http: {status}")
    print(res.desired_instances)
    return 0



#!/usr/bin/env python
from paasta_tools.kubernetes.application.controller_wrappers import (
    get_application_wrapper,
)
from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.kubernetes_tools import KubernetesDeploymentConfig
from paasta_tools.paasta_service_config_loader import PaastaServiceConfigLoader
from paasta_tools.utils import get_services_for_cluster
from paasta_tools.utils import load_system_paasta_config


def main() -> None:
    system_paasta_config = load_system_paasta_config()

    kube_client = KubeClient()

    services = {
        service
        for service, instance in get_services_for_cluster(
            cluster=system_paasta_config.get_cluster(), instance_type="kubernetes"
        )
    }

    for service in services:
        pscl = PaastaServiceConfigLoader(service=service, load_deployments=False)
        for instance_config in pscl.instance_configs(
            cluster=system_paasta_config.get_cluster(),
            instance_type_class=KubernetesDeploymentConfig,
        ):
            max_instances = instance_config.get_max_instances()
            if max_instances is not None:
                formatted_application = instance_config.format_kubernetes_app()
                formatted_application.spec.replicas = max_instances
                wrapper = get_application_wrapper(formatted_application)
                wrapper.soa_config = instance_config
                print(f"Scaling up {service}.{instance_config.instance}")
                wrapper.update(kube_client)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Usage: ./list_kubernetes_service_instances.py [options]

Enumerates all kubernetes instances for services in the soa directory that
are for the current cluster (defined by the kubernetes configuration file).

Outputs (to stdout) a list of service.instance (one per line)
for each instance found in kubernetes-<CLUSTER>.yaml for every folder
in the SOA Configuration directory.

Command line options:

- -d <SOA_DIR>, --soa-dir <SOA_DIR>: Specify a SOA config dir to read from
- -c <CLUSTER>, --cluster <CLUSTER>: Specify which cluster of services to read
"""
import argparse
import random
import sys
from typing import List

from paasta_tools import kubernetes_tools
from paasta_tools.utils import compose_job_id
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import get_services_for_cluster


def parse_args():
    parser = argparse.ArgumentParser(
        description="Lists kubernetes instances for a service."
    )
    parser.add_argument(
        "-c",
        "--cluster",
        dest="cluster",
        metavar="CLUSTER",
        default=None,
        help="define a specific cluster to read from",
    )
    parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    parser.add_argument(
        "--sanitise",
        action="store_true",
        help=(
            "Whether or not to sanitise service instance names before displaying "
            "them. Kubernetes apps created by PaaSTA use sanitised names."
        ),
    )
    parser.add_argument(
        "-t",
        "--instance-type",
        dest="instance_type",
        default="kubernetes",
        help="Instance type to list, default %(default)s",
    )
    parser.add_argument(
        "--shuffle",
        action="store_true",
        help="Shuffle the instances output.",
    )
    parser.add_argument(
        "--group-lines",
        type=int,
        dest="group_lines",
        help=(
            "Groups instances output into a desired number of lines with each instance separated by a space"
        ),
    )
    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    soa_dir = args.soa_dir
    cluster = args.cluster
    instances = get_services_for_cluster(
        cluster=cluster, instance_type=args.instance_type, soa_dir=soa_dir
    )
    service_instances = []
    for name, instance in instances:
        if args.sanitise:
            app_name = kubernetes_tools.get_kubernetes_app_name(name, instance)
        else:
            app_name = compose_job_id(name, instance)
        service_instances.append(app_name)
    if args.shuffle:
        random.shuffle(service_instances)

    if args.group_lines:
        group_lines(service_instances, args.group_lines)
    else:
        print("\n".join(service_instances))
    sys.exit(0)


def group_lines(service_instances: List[str], num_lines: int) -> None:
    output: List[List[str]] = [[] for _ in range(num_lines)]
    curr_index = 0
    for instance in service_instances:
        output[curr_index].append(instance)
        curr_index = (curr_index + 1) % num_lines

    for group in output:
        print(" ".join(group))


if __name__ == "__main__":
    main()



import copy
import logging
import os
import socket
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Type

import service_configuration_lib

from paasta_tools.autoscaling.utils import AutoscalingParamsDict
from paasta_tools.autoscaling.utils import MetricsProviderDict
from paasta_tools.paasta_service_config_loader import PaastaServiceConfigLoader
from paasta_tools.utils import BranchDictV2
from paasta_tools.utils import compose_job_id
from paasta_tools.utils import decompose_job_id
from paasta_tools.utils import deep_merge_dictionaries
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import DeployBlacklist
from paasta_tools.utils import DeployWhitelist
from paasta_tools.utils import InstanceConfig
from paasta_tools.utils import InstanceConfigDict
from paasta_tools.utils import InvalidInstanceConfig
from paasta_tools.utils import InvalidJobNameError
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import SystemPaastaConfig

log = logging.getLogger(__name__)
logging.getLogger("long_running_service_tools").setLevel(logging.WARNING)

ZK_PAUSE_AUTOSCALE_PATH = "/autoscaling/paused"
DEFAULT_CONTAINER_PORT = 8888

DEFAULT_AUTOSCALING_SETPOINT = 0.8
DEFAULT_DESIRED_ACTIVE_REQUESTS_PER_REPLICA = 1
DEFAULT_ACTIVE_REQUESTS_AUTOSCALING_MOVING_AVERAGE_WINDOW = 1800
DEFAULT_UWSGI_AUTOSCALING_MOVING_AVERAGE_WINDOW = 1800
DEFAULT_PISCINA_AUTOSCALING_MOVING_AVERAGE_WINDOW = 1800
DEFAULT_GUNICORN_AUTOSCALING_MOVING_AVERAGE_WINDOW = 1800

METRICS_PROVIDER_CPU = "cpu"
METRICS_PROVIDER_UWSGI = "uwsgi"
METRICS_PROVIDER_UWSGI_V2 = "uwsgi-v2"
METRICS_PROVIDER_GUNICORN = "gunicorn"
METRICS_PROVIDER_PISCINA = "piscina"
METRICS_PROVIDER_ACTIVE_REQUESTS = "active-requests"
METRICS_PROVIDER_PROMQL = "arbitrary_promql"

ALL_METRICS_PROVIDERS = [
    METRICS_PROVIDER_CPU,
    METRICS_PROVIDER_UWSGI,
    METRICS_PROVIDER_UWSGI_V2,
    METRICS_PROVIDER_GUNICORN,
    METRICS_PROVIDER_PISCINA,
    METRICS_PROVIDER_ACTIVE_REQUESTS,
    METRICS_PROVIDER_PROMQL,
]


class LongRunningServiceConfigDict(InstanceConfigDict, total=False):
    autoscaling: AutoscalingParamsDict
    drain_method: str
    fs_group: int
    container_port: int
    drain_method_params: Dict
    healthcheck_cmd: str
    healthcheck_grace_period_seconds: float
    healthcheck_interval_seconds: float
    healthcheck_max_consecutive_failures: int
    healthcheck_mode: str
    healthcheck_timeout_seconds: float
    healthcheck_uri: str
    instances: int
    max_instances: int
    min_instances: int
    nerve_ns: str
    network_mode: str
    registrations: List[str]
    replication_threshold: int
    bounce_start_deadline: float
    bounce_margin_factor: float
    should_ping_for_unhealthy_pods: bool
    weight: int


class ServiceNamespaceConfig(dict):
    def get_healthcheck_mode(self) -> str:
        """Get the healthcheck mode for the service. In most cases, this will match the mode
        of the service, but we do provide the opportunity for users to specify both. Default to the mode
        if no healthcheck_mode is specified.
        """
        healthcheck_mode = self.get("healthcheck_mode", None)
        if not healthcheck_mode:
            return self.get_mode()
        else:
            return healthcheck_mode

    def get_mode(self) -> str:
        """Get the mode that the service runs in and check that we support it.
        If the mode is not specified, we check whether the service uses smartstack
        in order to determine the appropriate default value. If proxy_port is specified
        in the config, the service uses smartstack, and we can thus safely assume its mode is http.
        If the mode is not defined and the service does not use smartstack, we set the mode to None.
        """
        mode = self.get("mode", None)
        if mode is None:
            if not self.is_in_smartstack():
                return None
            else:
                return "http"
        elif mode in ["http", "tcp", "https"]:
            return mode
        else:
            raise InvalidSmartstackMode("Unknown mode: %s" % mode)

    def get_healthcheck_uri(self) -> str:
        return self.get("healthcheck_uri", "/status")

    def get_discover(self) -> str:
        return self.get("discover", "region")

    def is_in_smartstack(self) -> bool:
        return "proxy_port" in self

    def get_timeout_server_ms(self) -> int:
        return self.get("timeout_server_ms", 1000)

    def get_longest_timeout_ms(self) -> int:
        """Calculate the longest amount of time a connection to this service might stay open."""
        return max(
            [self.get_timeout_server_ms()]
            + list(self.get("endpoint_timeouts", {}).values())
        )


class LongRunningServiceConfig(InstanceConfig):
    config_dict: LongRunningServiceConfigDict

    def __init__(
        self,
        service: str,
        cluster: str,
        instance: str,
        config_dict: LongRunningServiceConfigDict,
        branch_dict: Optional[BranchDictV2],
        soa_dir: str = DEFAULT_SOA_DIR,
    ) -> None:
        super().__init__(
            cluster=cluster,
            instance=instance,
            service=service,
            config_dict=config_dict,
            branch_dict=branch_dict,
            soa_dir=soa_dir,
        )

    def get_bounce_method(self) -> str:
        raise NotImplementedError

    def get_kubernetes_namespace(self) -> str:
        """
        Only needed on kubernetes LongRunningServiceConfig
        """
        raise NotImplementedError

    def get_sanitised_deployment_name(self) -> str:
        """
        Only needed on kubernetes LongRunningServiceConfig
        """
        raise NotImplementedError

    def get_service_name_smartstack(self) -> str:
        """
        This is just the service name here
        For cassandra we have to override this to support apollo
        """
        return self.get_service()

    def get_env(
        self, system_paasta_config: Optional[SystemPaastaConfig] = None
    ) -> Dict[str, str]:
        env = super().get_env(system_paasta_config=system_paasta_config)
        env["PAASTA_PORT"] = str(self.get_container_port())
        return env

    def get_container_port(self) -> int:
        return self.config_dict.get("container_port", DEFAULT_CONTAINER_PORT)

    def get_drain_method(self, service_namespace_config: ServiceNamespaceConfig) -> str:
        """Get the drain method specified in the service's configuration.

        :param service_config: The service instance's configuration dictionary
        :returns: The drain method specified in the config, or 'noop' if not specified"""
        default = "noop"
        # Default to hacheck draining if the service is in smartstack
        if service_namespace_config.is_in_smartstack():
            default = "hacheck"
        return self.config_dict.get("drain_method", default)

    def get_drain_method_params(
        self, service_namespace_config: ServiceNamespaceConfig
    ) -> Dict:
        """Get the drain method parameters specified in the service's configuration.

        :param service_config: The service instance's configuration dictionary
        :returns: The drain_method_params dictionary specified in the config, or {} if not specified"""
        default: Dict = {}
        if service_namespace_config.is_in_smartstack():
            default = {"delay": 60}
        return self.config_dict.get("drain_method_params", default)

    # FIXME(jlynch|2016-08-02, PAASTA-4964): DEPRECATE nerve_ns and remove it
    def get_nerve_namespace(self) -> str:
        return decompose_job_id(self.get_registrations()[0])[1]

    def get_registrations(self) -> List[str]:
        for registration in self.get_invalid_registrations():
            log.error(
                "Provided registration {} for service "
                "{} is invalid".format(registration, self.service)
            )

        registrations = self.config_dict.get("registrations", [])

        # Backwards compatibility with nerve_ns
        # FIXME(jlynch|2016-08-02, PAASTA-4964): DEPRECATE nerve_ns and remove it
        if not registrations and "nerve_ns" in self.config_dict:
            registrations.append(
                compose_job_id(self.service, self.config_dict["nerve_ns"])
            )

        return registrations or [compose_job_id(self.service, self.instance)]

    def get_invalid_registrations(self) -> List[str]:
        registrations = self.config_dict.get("registrations", [])
        invalid_registrations: List[str] = []
        for registration in registrations:
            try:
                decompose_job_id(registration)
            except InvalidJobNameError:
                invalid_registrations.append(registration)
        return invalid_registrations

    def get_replication_crit_percentage(self) -> int:
        return self.config_dict.get("replication_threshold", 50)

    def get_fs_group(self) -> Optional[int]:
        return self.config_dict.get("fs_group")

    def get_healthcheck_uri(
        self, service_namespace_config: ServiceNamespaceConfig
    ) -> str:
        return self.config_dict.get(
            "healthcheck_uri", service_namespace_config.get_healthcheck_uri()
        )

    def get_healthcheck_cmd(self) -> str:
        cmd = self.config_dict.get("healthcheck_cmd", None)
        if cmd is None:
            raise InvalidInstanceConfig(
                "healthcheck mode 'cmd' requires a healthcheck_cmd to run"
            )
        else:
            return cmd

    def get_healthcheck_grace_period_seconds(self) -> float:
        """
        How long before kubernetes will start sending healthcheck and liveness probes.
        """
        return self.config_dict.get("healthcheck_grace_period_seconds", 60)

    def get_healthcheck_interval_seconds(self) -> float:
        return self.config_dict.get("healthcheck_interval_seconds", 10)

    def get_healthcheck_timeout_seconds(self) -> float:
        return self.config_dict.get("healthcheck_timeout_seconds", 10)

    def get_healthcheck_max_consecutive_failures(self) -> int:
        return self.config_dict.get("healthcheck_max_consecutive_failures", 30)

    def get_healthcheck_mode(
        self, service_namespace_config: ServiceNamespaceConfig
    ) -> str:
        mode = self.config_dict.get("healthcheck_mode", None)
        if mode is None:
            mode = service_namespace_config.get_healthcheck_mode()
        elif mode not in ["http", "https", "tcp", "cmd", None]:
            raise InvalidHealthcheckMode("Unknown mode: %s" % mode)
        return mode

    def get_bounce_start_deadline(self) -> float:
        return self.config_dict.get("bounce_start_deadline", 0)

    def get_autoscaled_instances(self) -> int:
        raise NotImplementedError()

    def get_instances(self, with_limit: bool = True) -> int:
        """Gets the number of instances for a service, ignoring whether the user has requested
        the service to be started or stopped"""
        if self.is_autoscaling_enabled():
            autoscaled_instances = self.get_autoscaled_instances()
            if autoscaled_instances is None:
                return self.get_max_instances()
            else:
                limited_instances = (
                    self.limit_instance_count(autoscaled_instances)
                    if with_limit
                    else autoscaled_instances
                )
                return limited_instances
        else:
            instances = self.config_dict.get("instances", 1)
            log.debug("Autoscaling not enabled, returning %d instances" % instances)
            return instances

    def get_min_instances(self) -> int:
        return self.config_dict.get("min_instances", 1)

    def is_autoscaling_enabled(self) -> bool:
        return self.get_max_instances() is not None

    def get_max_instances(self) -> Optional[int]:
        return self.config_dict.get("max_instances", None)

    def get_desired_instances(self) -> int:
        """Get the number of instances specified in zookeeper or the service's configuration.
        If the number of instances in zookeeper is less than min_instances, returns min_instances.
        If the number of instances in zookeeper is greater than max_instances, returns max_instances.

        Defaults to 0 if not specified in the config.

        :returns: The number of instances specified in the config, 0 if not
                  specified or if desired_state is not 'start'.
        """
        if self.get_desired_state() == "start":
            return self.get_instances()
        else:
            log.debug("Instance is set to stop. Returning '0' instances")
            return 0

    def limit_instance_count(self, instances: int) -> int:
        """
        Returns param instances if it is between min_instances and max_instances.
        Returns max_instances if instances > max_instances
        Returns min_instances if instances < min_instances
        """
        return max(self.get_min_instances(), min(self.get_max_instances(), instances))

    def get_autoscaling_params(self) -> AutoscalingParamsDict:
        default_provider_params: MetricsProviderDict = {
            "type": METRICS_PROVIDER_CPU,
            "decision_policy": "proportional",
            "setpoint": DEFAULT_AUTOSCALING_SETPOINT,
        }

        params = copy.deepcopy(
            self.config_dict.get("autoscaling", AutoscalingParamsDict({}))
        )
        if "metrics_providers" not in params or len(params["metrics_providers"]) == 0:
            params["metrics_providers"] = [default_provider_params]
        else:
            params["metrics_providers"] = [
                deep_merge_dictionaries(
                    overrides=provider,
                    defaults=default_provider_params,
                )
                for provider in params["metrics_providers"]
            ]
        return params

    def get_autoscaling_metrics_provider(
        self, provider_type: str
    ) -> Optional[MetricsProviderDict]:
        autoscaling_params = self.get_autoscaling_params()
        # We only allow one metric provider of each type, so we can bail early if we find a match
        for provider in autoscaling_params["metrics_providers"]:
            if provider["type"] == provider_type:
                return provider
        return None

    def should_use_metrics_provider(self, provider_type: str) -> bool:
        return (
            self.is_autoscaling_enabled()
            and self.get_autoscaling_metrics_provider(provider_type) is not None
        )

    def validate(
        self,
        params: Optional[List[str]] = None,
    ) -> List[str]:
        error_messages = super().validate(params=params)
        invalid_registrations = self.get_invalid_registrations()
        if invalid_registrations:
            service_instance = compose_job_id(self.service, self.instance)
            registrations_str = ", ".join(invalid_registrations)
            error_messages.append(
                f"Service registrations must be of the form service.registration. "
                f"The following registrations for {service_instance} are "
                f"invalid: {registrations_str}"
            )
        return error_messages

    def get_bounce_margin_factor(self) -> float:
        return self.config_dict.get("bounce_margin_factor", 0.95)

    def get_should_ping_for_unhealthy_pods(self, default: bool) -> bool:
        return self.config_dict.get("should_ping_for_unhealthy_pods", default)

    def get_weight(self) -> int:
        return self.config_dict.get("weight", 10)


class InvalidHealthcheckMode(Exception):
    pass


def get_healthcheck_for_instance(
    service: str,
    instance: str,
    service_manifest: LongRunningServiceConfig,
    random_port: int,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> Tuple[Optional[str], Optional[str]]:
    """
    Returns healthcheck for a given service instance in the form of a tuple (mode, healthcheck_command)
    or (None, None) if no healthcheck
    """
    namespace = service_manifest.get_nerve_namespace()
    smartstack_config = load_service_namespace_config(
        service=service, namespace=namespace, soa_dir=soa_dir
    )
    mode = service_manifest.get_healthcheck_mode(smartstack_config)
    hostname = socket.getfqdn()

    if mode == "http" or mode == "https":
        path = service_manifest.get_healthcheck_uri(smartstack_config)
        healthcheck_command = "%s://%s:%d%s" % (mode, hostname, random_port, path)
    elif mode == "tcp":
        healthcheck_command = "%s://%s:%d" % (mode, hostname, random_port)
    elif mode == "cmd":
        healthcheck_command = service_manifest.get_healthcheck_cmd()
    else:
        mode = None
        healthcheck_command = None
    return (mode, healthcheck_command)


def load_service_namespace_config(
    service: str, namespace: str, soa_dir: str = DEFAULT_SOA_DIR
) -> ServiceNamespaceConfig:
    """Attempt to read the configuration for a service's namespace in a more strict fashion.

    Retrieves the following keys:

    - proxy_port: the proxy port defined for the given namespace
    - healthcheck_mode: the mode for the healthcheck (http or tcp)
    - healthcheck_port: An alternate port to use for health checking
    - healthcheck_uri: URI target for healthchecking
    - healthcheck_timeout_s: healthcheck timeout in seconds
    - healthcheck_body_expect: an expected string in healthcheck response body
    - updown_timeout_s: updown_service timeout in seconds
    - timeout_connect_ms: proxy frontend timeout in milliseconds
    - timeout_server_ms: proxy server backend timeout in milliseconds
    - retries: the number of retries on a proxy backend
    - mode: the mode the service is run in (http or tcp)
    - routes: a list of tuples of (source, destination)
    - discover: the scope at which to discover services e.g. 'habitat'
    - advertise: a list of scopes to advertise services at e.g. ['habitat', 'region']
    - extra_advertise: a list of tuples of (source, destination)
      e.g. [('region:dc6-prod', 'region:useast1-prod')]
    - extra_healthcheck_headers: a dict of HTTP headers that must
      be supplied when health checking. E.g. { 'Host': 'example.com' }
    - lb_policy: Envoy load balancer policies. E.g. "ROUND_ROBIN"

    :param service: The service name
    :param namespace: The namespace to read
    :param soa_dir: The SOA config directory to read from
    :returns: A dict of the above keys, if they were defined
    """

    smartstack_config = service_configuration_lib.read_extra_service_information(
        service_name=service,
        extra_info="smartstack",
        soa_dir=soa_dir,
        deepcopy=False,
    )

    namespace_config_from_file = smartstack_config.get(namespace, {})

    service_namespace_config = ServiceNamespaceConfig()
    # We can't really use .get, as we don't want the key to be in the returned
    # dict at all if it doesn't exist in the config file.
    # We also can't just copy the whole dict, as we only care about some keys
    # and there's other things that appear in the smartstack section in
    # several cases.
    key_whitelist = {
        "healthcheck_mode",
        "healthcheck_uri",
        "healthcheck_port",
        "healthcheck_timeout_s",
        "healthcheck_body_expect",
        "updown_timeout_s",
        "proxy_port",
        "timeout_connect_ms",
        "timeout_server_ms",
        "retries",
        "mode",
        "discover",
        "advertise",
        "extra_healthcheck_headers",
        "lb_policy",
        "endpoint_timeouts",
    }

    for key, value in namespace_config_from_file.items():
        if key in key_whitelist:
            service_namespace_config[key] = value

    # Other code in paasta_tools checks 'mode' after the config file
    # is loaded, so this ensures that it is set to the appropriate default
    # if not otherwise specified, even if appropriate default is None.
    service_namespace_config["mode"] = service_namespace_config.get_mode()

    if "routes" in namespace_config_from_file:
        service_namespace_config["routes"] = [
            (route["source"], dest)
            for route in namespace_config_from_file["routes"]
            for dest in route["destinations"]
        ]

    if "extra_advertise" in namespace_config_from_file:
        service_namespace_config["extra_advertise"] = [
            (src, dst)
            for src in namespace_config_from_file["extra_advertise"]
            for dst in namespace_config_from_file["extra_advertise"][src]
        ]

    return service_namespace_config


class InvalidSmartstackMode(Exception):
    pass


def get_proxy_port_for_instance(
    service_config: LongRunningServiceConfig,
) -> Optional[int]:
    """Get the proxy_port defined in the first namespace configuration for a
    service instance.

    This means that the namespace first has to be loaded from the service instance's
    configuration, and then the proxy_port has to loaded from the smartstack configuration
    for that namespace.

    :param service_config: The instance of the services LongRunningServiceConfig
    :returns: The proxy_port for the service instance, or None if not defined"""
    registration = service_config.get_registrations()[0]
    service, namespace, _, __ = decompose_job_id(registration)
    nerve_dict = load_service_namespace_config(
        service=service, namespace=namespace, soa_dir=service_config.soa_dir
    )
    return nerve_dict.get("proxy_port")


def host_passes_blacklist(
    host_attributes: Mapping[str, str], blacklist: DeployBlacklist
) -> bool:
    """
    :param host: A single host attributes dict
    :param blacklist: A list of lists like [["location_type", "location"], ["foo", "bar"]]
    :returns: boolean, True if the host gets passed the blacklist
    """
    try:
        for location_type, location in blacklist:
            if host_attributes.get(location_type) == location:
                return False
    except ValueError as e:
        log.error(f"Errors processing the following blacklist: {blacklist}")
        log.error("I will assume the host does not pass\nError was: %s" % e)
        return False
    return True


def host_passes_whitelist(
    host_attributes: Mapping[str, str], whitelist: DeployWhitelist
) -> bool:
    """
    :param host: A single host attributes dict.
    :param whitelist: A 2 item list like ["location_type", ["location1", 'location2']]
    :returns: boolean, True if the host gets past the whitelist
    """
    # No whitelist, so disable whitelisting behavior.
    if whitelist is None or len(whitelist) == 0:
        return True
    try:
        (location_type, locations) = whitelist
        if host_attributes.get(location_type) in locations:
            return True
    except ValueError as e:
        log.error(f"Errors processing the following whitelist: {whitelist}")
        log.error("I will assume the host does not pass\nError was: %s" % e)
        return False
    return False


def get_all_namespaces(
    soa_dir: str = DEFAULT_SOA_DIR,
) -> Sequence[Tuple[str, ServiceNamespaceConfig]]:
    """Get all the smartstack namespaces across all services.
    This is mostly so synapse can get everything it needs in one call.

    :param soa_dir: The SOA config directory to read from
    :returns: A list of tuples of the form (service.namespace, namespace_config)"""
    rootdir = os.path.abspath(soa_dir)
    namespace_list: List[Tuple[str, ServiceNamespaceConfig]] = []
    for srv_dir in os.listdir(rootdir):
        namespace_list.extend(get_all_namespaces_for_service(srv_dir, soa_dir))
    return namespace_list


def get_all_namespaces_for_service(
    service: str, soa_dir: str = DEFAULT_SOA_DIR, full_name: bool = True
) -> Sequence[Tuple[str, ServiceNamespaceConfig]]:
    """Get all the smartstack namespaces listed for a given service name.

    :param service: The service name
    :param soa_dir: The SOA config directory to read from
    :param full_name: A boolean indicating if the service name should be prepended to the namespace in the
                      returned tuples as described below (Default: True)
    :returns: A list of tuples of the form (service<SPACER>namespace, namespace_config) if full_name is true,
              otherwise of the form (namespace, namespace_config)
    """
    service_config = service_configuration_lib.read_service_configuration(
        service, soa_dir
    )
    smartstack = service_config.get("smartstack", {})
    namespace_list = []
    for namespace in smartstack:
        if full_name:
            name = compose_job_id(service, namespace)
        else:
            name = namespace
        namespace_list.append((name, smartstack[namespace]))
    return namespace_list


def get_expected_instance_count_for_namespace(
    service: str,
    namespace: str,
    instance_type_class: Type[LongRunningServiceConfig],
    cluster: str = None,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> int:
    """Get the number of expected instances for a namespace, based on the number
    of instances set to run on that namespace as specified in service configuration files.

    :param service: The service's name
    :param namespace: The namespace for that service to check
    instance_type_class: The type of the instance, options are e.g. KubernetesDeploymentConfig,
    :param soa_dir: The SOA configuration directory to read from
    :returns: An integer value of the # of expected instances for the namespace"""
    total_expected = 0
    if not cluster:
        cluster = load_system_paasta_config().get_cluster()

    pscl = PaastaServiceConfigLoader(
        service=service, soa_dir=soa_dir, load_deployments=False
    )
    for job_config in pscl.instance_configs(
        cluster=cluster, instance_type_class=instance_type_class
    ):
        if f"{service}.{namespace}" in job_config.get_registrations():
            total_expected += job_config.get_instances()
    return total_expected



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from paasta_tools.cli.utils import get_instance_configs_for_service
from paasta_tools.cli.utils import lazy_choices_completer
from paasta_tools.cli.utils import validate_service_name
from paasta_tools.spark_tools import SPARK_EXECUTOR_NAMESPACE
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import list_clusters
from paasta_tools.utils import list_services


def add_subparser(subparsers) -> None:
    list_parser = subparsers.add_parser(
        "list-namespaces",
        help="Lists all k8s namespaces used by instances of a service",
    )
    list_parser.add_argument(
        "-s",
        "--service",
        help="Name of the service which you want to list the namespaces for.",
        required=True,
    ).completer = lazy_choices_completer(list_services)
    # Most services likely don't need to filter by cluster/instance, and can add namespaces from all instances
    list_parser.add_argument(
        "-i",
        "--instance",
        help="Instance of the service that you want to list namespaces for. Like 'main' or 'canary'.",
        required=False,
    )
    list_parser.add_argument(
        "-c",
        "--cluster",
        help="Clusters that you want to list namespaces for. Like 'pnw-prod' or 'norcal-stagef'.",
        required=False,
    ).completer = lazy_choices_completer(list_clusters)
    list_parser.add_argument(
        "-y",
        "-d",
        "--soa-dir",
        dest="soa_dir",
        default=DEFAULT_SOA_DIR,
        required=False,
        help="define a different soa config directory",
    )
    list_parser.set_defaults(command=paasta_list_namespaces)


def paasta_list_namespaces(args):
    service = args.service
    soa_dir = args.soa_dir
    validate_service_name(service, soa_dir)

    namespaces = set()
    instance_configs = get_instance_configs_for_service(
        service=service, soa_dir=soa_dir, instances=args.instance, clusters=args.cluster
    )
    for instance in instance_configs:
        # We skip non-k8s instance types
        if instance.get_instance_type() in ("paasta-native", "adhoc"):
            continue
        namespaces.add(instance.get_namespace())
        # Tron instances are TronActionConfigs
        if (
            instance.get_instance_type() == "tron"
            and instance.get_executor() == "spark"
        ):
            # We also need paasta-spark for spark executors
            namespaces.add(SPARK_EXECUTOR_NAMESPACE)

    # Print in list format to be used in iam_roles
    print(list(namespaces))
    return 0



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
import collections
import csv
import logging
import random
from typing import Any
from typing import cast
from typing import Collection
from typing import DefaultDict
from typing import Dict
from typing import Iterable
from typing import List
from typing import MutableMapping
from typing import NamedTuple
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import TypeVar
from typing import Union

import requests
from kubernetes.client import V1Node
from kubernetes.client import V1Pod
from mypy_extensions import TypedDict

from paasta_tools import envoy_tools
from paasta_tools import kubernetes_tools
from paasta_tools import long_running_service_tools
from paasta_tools import mesos_tools
from paasta_tools.long_running_service_tools import LongRunningServiceConfig
from paasta_tools.mesos.exceptions import NoSlavesAvailableError
from paasta_tools.monitoring_tools import ReplicationChecker
from paasta_tools.utils import compose_job_id
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import DeployBlacklist
from paasta_tools.utils import get_user_agent
from paasta_tools.utils import SystemPaastaConfig


class HaproxyBackend(TypedDict, total=False):
    check_code: str
    check_duration: str
    check_status: str
    lastchg: str
    pxname: str
    svname: str
    status: str


log = logging.getLogger(__name__)


def retrieve_haproxy_csv(
    synapse_host: str, synapse_port: int, synapse_haproxy_url_format: str, scope: str
) -> Iterable[Dict[str, str]]:
    """Retrieves the haproxy csv from the haproxy web interface

    :param synapse_host: A host that this check should contact for replication information.
    :param synapse_port: A integer that this check should contact for replication information.
    :param synapse_haproxy_url_format: The format of the synapse haproxy URL.
    :param scope: scope
    :returns reader: a csv.DictReader object
    """
    synapse_uri = synapse_haproxy_url_format.format(
        host=synapse_host, port=synapse_port, scope=scope
    )

    # timeout after 1 second and retry 3 times
    haproxy_request = requests.Session()
    haproxy_request.headers.update({"User-Agent": get_user_agent()})
    haproxy_request.mount("http://", requests.adapters.HTTPAdapter(max_retries=3))
    haproxy_request.mount("https://", requests.adapters.HTTPAdapter(max_retries=3))
    haproxy_response = haproxy_request.get(synapse_uri, timeout=1)
    haproxy_data = haproxy_response.text
    reader = csv.DictReader(haproxy_data.splitlines())
    return reader


def get_backends(
    service: str, synapse_host: str, synapse_port: int, synapse_haproxy_url_format: str
) -> List[HaproxyBackend]:
    """Fetches the CSV from haproxy and returns a list of backends,
    regardless of their state.

    :param service: If None, return backends for all services, otherwise only return backends for this particular
                    service.
    :param synapse_host: A host that this check should contact for replication information.
    :param synapse_port: A integer that this check should contact for replication information.
    :param synapse_haproxy_url_format: The format of the synapse haproxy URL.
    :returns backends: A list of dicts representing the backends of all
                       services or the requested service
    """
    if service:
        services = [service]
    else:
        services = None
    return get_multiple_backends(
        services,
        synapse_host=synapse_host,
        synapse_port=synapse_port,
        synapse_haproxy_url_format=synapse_haproxy_url_format,
    )


def get_multiple_backends(
    services: Optional[Collection[str]],
    synapse_host: str,
    synapse_port: int,
    synapse_haproxy_url_format: str,
) -> List[HaproxyBackend]:
    """Fetches the CSV from haproxy and returns a list of backends,
    regardless of their state.

    :param services: If None, return backends for all services, otherwise only return backends for these particular
                     services.
    :param synapse_host: A host that this check should contact for replication information.
    :param synapse_port: A integer that this check should contact for replication information.
    :param synapse_haproxy_url_format: The format of the synapse haproxy URL.
    :returns backends: A list of dicts representing the backends of all
                       services or the requested service
    """

    if services is not None and len(services) == 1:
        (scope,) = services
    else:
        # Maybe if there's like two or three services we could make two queries, or find the longest common substring.
        # For now let's just hope this is rare and fetch all data.
        scope = ""

    reader = retrieve_haproxy_csv(
        synapse_host,
        synapse_port,
        synapse_haproxy_url_format=synapse_haproxy_url_format,
        scope=scope,
    )
    backends = []

    for line in reader:
        # clean up two irregularities of the CSV output, relative to
        # DictReader's behavior there's a leading "# " for no good reason:
        line["pxname"] = line.pop("# pxname")
        # and there's a trailing comma on every line:
        line.pop("")

        # Look for the service in question and ignore the fictional
        # FRONTEND/BACKEND hosts, use starts_with so that hosts that are UP
        # with 1/X healthchecks to go before going down get counted as UP:
        ha_slave, ha_service = line["svname"], line["pxname"]
        if (services is None or ha_service in services) and ha_slave not in (
            "FRONTEND",
            "BACKEND",
        ):
            backends.append(cast(HaproxyBackend, line))

    return backends


def load_smartstack_info_for_service(
    service: str,
    namespace: str,
    blacklist: DeployBlacklist,
    system_paasta_config: SystemPaastaConfig,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> Dict[str, Dict[str, int]]:
    """Retrieves number of available backends for given service

    :param service: A service name
    :param namespace: A Smartstack namespace
    :param blacklist: A list of blacklisted location tuples in the form (location, value)
    :param system_paasta_config: A SystemPaastaConfig object representing the system configuration.
    :param soa_dir: SOA dir
    :returns: a dictionary of the form

    ::

        {
          'location_type': {
              'unique_location_name': {
                  'service.instance': <# ofavailable backends>
              },
              'other_unique_location_name': ...
          }
        }

    """
    service_namespace_config = long_running_service_tools.load_service_namespace_config(
        service=service, namespace=namespace, soa_dir=soa_dir
    )
    discover_location_type = service_namespace_config.get_discover()
    return get_smartstack_replication_for_attribute(
        attribute=discover_location_type,
        service=service,
        namespace=namespace,
        blacklist=blacklist,
        system_paasta_config=system_paasta_config,
    )


def get_smartstack_replication_for_attribute(
    attribute: str,
    service: str,
    namespace: str,
    blacklist: DeployBlacklist,
    system_paasta_config: SystemPaastaConfig,
) -> Dict[str, Dict[str, int]]:
    """Loads smartstack replication from a host with the specified attribute

    :param attribute: a Mesos attribute
    :param service: A service name, like 'example_service'
    :param namespace: A particular smartstack namespace to inspect, like 'main'
    :param blacklist: A list of blacklisted location tuples in the form of (location, value)
    :param system_paasta_config: A SystemPaastaConfig object representing the system configuration.
    :returns: a dictionary of the form {'<unique_attribute_value>': <smartstack replication hash>}
              (the dictionary will contain keys for unique all attribute values)
    """
    replication_info = {}
    filtered_slaves = mesos_tools.get_all_slaves_for_blacklist_whitelist(
        blacklist=blacklist, whitelist=None
    )
    if not filtered_slaves:
        raise NoSlavesAvailableError

    attribute_slave_dict = mesos_tools.get_mesos_slaves_grouped_by_attribute(
        slaves=filtered_slaves, attribute=attribute
    )

    full_name = compose_job_id(service, namespace)

    for value, hosts in attribute_slave_dict.items():
        # arbitrarily choose the first host with a given attribute to query for replication stats
        synapse_host = hosts[0]["hostname"]
        repl_info = get_replication_for_services(
            synapse_host=synapse_host,
            synapse_port=system_paasta_config.get_synapse_port(),
            synapse_haproxy_url_format=system_paasta_config.get_synapse_haproxy_url_format(),
            services=[full_name],
        )
        replication_info[value] = repl_info

    return replication_info


def get_replication_for_all_services(
    synapse_host: str, synapse_port: int, synapse_haproxy_url_format: str
) -> Dict[str, int]:
    """Returns the replication level for all services known to this synapse haproxy

    :param synapse_host: The host that this check should contact for replication information.
    :param synapse_port: The port that this check should contact for replication information.
    :param synapse_haproxy_url_format: The format of the synapse haproxy URL.
    :returns available_instance_counts: A dictionary mapping the service names
                                        to an integer number of available replicas.
    """
    backends = get_multiple_backends(
        services=None,
        synapse_host=synapse_host,
        synapse_port=synapse_port,
        synapse_haproxy_url_format=synapse_haproxy_url_format,
    )
    return collections.Counter([b["pxname"] for b in backends if backend_is_up(b)])


def get_replication_for_services(
    synapse_host: str,
    synapse_port: int,
    synapse_haproxy_url_format: str,
    services: Collection[str],
) -> Dict[str, int]:
    """Returns the replication level for the provided services

    This check is intended to be used with an haproxy load balancer, and
    relies on the implementation details of that choice.

    :param synapse_host: The host that this check should contact for replication information.
    :param synapse_port: The port that this check should contact for replication information.
    :param synapse_haproxy_url_format: The format of the synapse haproxy URL.
    :param services: A list of strings that are the service names
                          that should be checked for replication.

    :returns available_instance_counts: A dictionary mapping the service names
                                  to an integer number of available
                                  replicas
    :returns None: If it cannot connect to the specified synapse host and port
    """
    backends = get_multiple_backends(
        services=services,
        synapse_host=synapse_host,
        synapse_port=synapse_port,
        synapse_haproxy_url_format=synapse_haproxy_url_format,
    )

    counter = collections.Counter([b["pxname"] for b in backends if backend_is_up(b)])
    return {sn: counter[sn] for sn in services}


def backend_is_up(backend: HaproxyBackend) -> bool:
    """Returns whether a server is receiving traffic in HAProxy.

    :param backend: backend dict, like one of those returned by smartstack_tools.get_multiple_backends.

    :returns is_up: Whether the backend is in a state that receives traffic.
    """
    return str(backend["status"]).startswith("UP")


def ip_port_hostname_from_svname(svname: str) -> Tuple[str, int, str]:
    """This parses the haproxy svname that smartstack creates.
    In old versions of synapse, this is in the format ip:port_hostname.
    In versions newer than dd5843c987740a5d5ce1c83b12b258b7253784a8 it is
    hostname_ip:port

    :param svname: A svname, in either of the formats described above
    :returns ip_port_hostname: A tuple of ip, port, hostname.
    """
    # split into parts
    parts = set(svname.split("_"))

    # find those that can be split by : - this is the ip:port
    # there will only be 1 of these
    ip_ports = {part for part in parts if len(part.split(":")) == 2}

    # the one *not* in the list is the hostname
    hostname = parts.difference(ip_ports).pop()

    ip, port = ip_ports.pop().split(":")
    return ip, int(port), hostname


def are_services_up_on_ip_port(
    synapse_host: str,
    synapse_port: int,
    synapse_haproxy_url_format: str,
    services: Collection[str],
    host_ip: str,
    host_port: int,
) -> bool:
    backends = get_multiple_backends(
        services,
        synapse_host=synapse_host,
        synapse_port=synapse_port,
        synapse_haproxy_url_format=synapse_haproxy_url_format,
    )
    backends_by_ip_port: DefaultDict[
        Tuple[str, int], List[HaproxyBackend]
    ] = collections.defaultdict(list)

    for backend in backends:
        ip, port, _ = ip_port_hostname_from_svname(backend["svname"])
        backends_by_ip_port[ip, port].append(backend)

    backends_on_ip = backends_by_ip_port[host_ip, host_port]
    # any backend being up is okay because a previous backend
    # may have had the same IP and synapse only removes them completely
    # after some time
    services_with_atleast_one_backend_up = {service: False for service in services}
    for service in services:
        for be in backends_on_ip:
            if be["pxname"] == service and backend_is_up(be):
                services_with_atleast_one_backend_up[service] = True
    return all(services_with_atleast_one_backend_up.values())


def match_backends_and_pods(
    backends: Iterable[HaproxyBackend], pods: Iterable[V1Pod]
) -> List[Tuple[Optional[HaproxyBackend], Optional[V1Pod]]]:
    """Returns tuples of matching (backend, pod) pairs, as matched by IP. Each backend will be listed exactly
    once. If a backend does not match with a pod, (backend, None) will be included.
    If a pod's IP does not match with any backends, (None, pod) will be included.

    :param backends: An iterable of haproxy backend dictionaries, e.g. the list returned by
                     smartstack_tools.get_multiple_backends.
    :param pods: An iterable of V1Pod objects.
    """

    # { ip : [backend1, backend2], ... }
    backends_by_ip: DefaultDict[str, List[HaproxyBackend]] = collections.defaultdict(
        list
    )
    backend_pod_pairs = []

    for backend in backends:
        ip, port, _ = ip_port_hostname_from_svname(backend["svname"])
        backends_by_ip[ip].append(backend)

    for pod in pods:
        ip = pod.status.pod_ip
        for backend in backends_by_ip.pop(ip, [None]):
            backend_pod_pairs.append((backend, pod))

    # we've been popping in the above loop, so anything left didn't match a k8s pod.
    for backends in backends_by_ip.values():
        for backend in backends:
            backend_pod_pairs.append((backend, None))

    return backend_pod_pairs


_MesosSlaveDict = TypeVar(
    "_MesosSlaveDict", bound=Dict
)  # no type has been defined in mesos_tools for these yet.


class DiscoveredHost(NamedTuple):
    hostname: str
    pool: str


class ServiceDiscoveryProvider(abc.ABC):

    NAME = "..."

    @abc.abstractmethod
    def get_replication_for_all_services(self, hostname: str) -> Dict[str, int]:
        ...


class SmartstackServiceDiscovery(ServiceDiscoveryProvider):

    NAME = "Smartstack"

    def __init__(self, system_paasta_config: SystemPaastaConfig) -> None:
        self._synapse_port = system_paasta_config.get_synapse_port()
        self._synapse_haproxy_url_format = (
            system_paasta_config.get_synapse_haproxy_url_format()
        )

    def get_replication_for_all_services(self, hostname: str) -> Dict[str, int]:
        return get_replication_for_all_services(
            synapse_host=hostname,
            synapse_port=self._synapse_port,
            synapse_haproxy_url_format=self._synapse_haproxy_url_format,
        )


class EnvoyServiceDiscovery(ServiceDiscoveryProvider):

    NAME = "Envoy"

    def __init__(self, system_paasta_config: SystemPaastaConfig) -> None:
        self._envoy_admin_port = system_paasta_config.get_envoy_admin_port()
        self._envoy_admin_endpoint_format = (
            system_paasta_config.get_envoy_admin_endpoint_format()
        )

    def get_replication_for_all_services(self, hostname: str) -> Dict[str, int]:
        return envoy_tools.get_replication_for_all_services(
            envoy_host=hostname,
            envoy_admin_port=self._envoy_admin_port,
            envoy_admin_endpoint_format=self._envoy_admin_endpoint_format,
        )


def get_service_discovery_providers(
    system_paasta_config: SystemPaastaConfig,
) -> List[ServiceDiscoveryProvider]:
    providers: List[ServiceDiscoveryProvider] = []
    for name, _ in system_paasta_config.get_service_discovery_providers().items():
        if name == "smartstack":
            providers.append(SmartstackServiceDiscovery(system_paasta_config))
        elif name == "envoy":
            providers.append(EnvoyServiceDiscovery(system_paasta_config))
        else:
            log.warn("unknown provider")
    return providers


class BaseReplicationChecker(ReplicationChecker):
    """Base class for checking replication. Extendable for different frameworks.

    Optimized for multiple queries. Gets the list of backends from service
    discovery provider only once per location and reuse it in all subsequent
    calls of BaseReplicationChecker.get_replication_for_instance().

    get_allowed_locations_and_hosts must be implemented in sub class

    A list of service discovery providers to collect information about
    instances and their status must be provided as
    `service_discovery_providers`.
    """

    def __init__(
        self,
        system_paasta_config: SystemPaastaConfig,
        service_discovery_providers: Iterable[ServiceDiscoveryProvider],
    ) -> None:
        self._system_paasta_config = system_paasta_config
        self._cache: Dict[Tuple[str, str], Dict[str, int]] = {}
        self._service_discovery_providers = service_discovery_providers

    @abc.abstractmethod
    def get_allowed_locations_and_hosts(
        self, instance_config: LongRunningServiceConfig
    ) -> Dict[str, Sequence[DiscoveredHost]]:
        ...

    def get_replication_for_instance(
        self, instance_config: LongRunningServiceConfig
    ) -> Dict[str, Dict[str, Dict[str, int]]]:
        """Returns the number of registered instances in each discoverable
        location for each service dicrovery provider.

        :param instance_config: An instance of LongRunningServiceConfig.
        :returns: a dict {'service_discovery_provider': {'location_type': {'service.instance': int}}}
        """
        replication_infos = {}
        for provider in self._service_discovery_providers:
            replication_info = {}
            attribute_host_dict = self.get_allowed_locations_and_hosts(instance_config)
            instance_pool = instance_config.get_pool()
            for location, hosts in attribute_host_dict.items():
                # Try to get information from all available hosts in the pool before giving up
                hostnames = self.get_hostnames_in_pool(hosts, instance_pool)
                for hostname in hostnames:
                    try:
                        replication_info[location] = self._get_replication_info(
                            location, hostname, instance_config, provider
                        )
                        break
                    except Exception as e:
                        log.warn(
                            f"Error while getting replication info for {location} from {hostname}: {e}"
                        )
                        if hostname == hostnames[-1]:
                            # Last hostname failed, giving up
                            raise
            replication_infos[provider.NAME] = replication_info
        return replication_infos

    def get_first_host_in_pool(self, hosts: Sequence[DiscoveredHost], pool: str) -> str:
        for host in hosts:
            if host.pool == pool:
                return host.hostname
        return hosts[0].hostname

    def get_hostname_in_pool(self, hosts: Sequence[DiscoveredHost], pool: str) -> str:
        return random.choice(self.get_hostnames_in_pool(hosts, pool))

    def get_hostnames_in_pool(
        self, hosts: Sequence[DiscoveredHost], pool: str
    ) -> Sequence[str]:
        hostnames = []
        for host in hosts:
            if host.pool == pool:
                hostnames.append(host.hostname)
        if len(hostnames) == 0:
            hostnames.append(hosts[0].hostname)
        return hostnames

    def _get_replication_info(
        self,
        location: str,
        hostname: str,
        instance_config: LongRunningServiceConfig,
        provider: ServiceDiscoveryProvider,
    ) -> Dict[str, int]:
        """Returns service.instance and the number of instances registered in smartstack
        at the location as a dict.

        :param location: A string that identifies a habitat, a region and etc.
        :param hostname: A mesos slave hostname to read replication information from.
        :param instance_config: An instance of LongRunningServiceConfig.
        :returns: A dict {"service.instance": number_of_instances}.
        """
        full_name = compose_job_id(instance_config.service, instance_config.instance)
        key = (location, provider.NAME)
        replication_info = self._cache.get(key)
        if replication_info is None:
            replication_info = provider.get_replication_for_all_services(hostname)
            self._cache[key] = replication_info
        return {full_name: replication_info[full_name]}


class KubeSmartstackEnvoyReplicationChecker(BaseReplicationChecker):
    def __init__(
        self, nodes: Sequence[V1Node], system_paasta_config: SystemPaastaConfig
    ) -> None:
        self.nodes = nodes
        super().__init__(
            system_paasta_config=system_paasta_config,
            service_discovery_providers=get_service_discovery_providers(
                system_paasta_config
            ),
        )

    def get_allowed_locations_and_hosts(
        self, instance_config: LongRunningServiceConfig
    ) -> Dict[str, Sequence[DiscoveredHost]]:
        discover_location_type = kubernetes_tools.load_service_namespace_config(
            service=instance_config.service,
            namespace=instance_config.get_nerve_namespace(),
            soa_dir=instance_config.soa_dir,
        ).get_discover()

        attribute_to_nodes = kubernetes_tools.get_nodes_grouped_by_attribute(
            nodes=self.nodes, attribute=discover_location_type
        )
        ret: Dict[str, Sequence[DiscoveredHost]] = {}
        for attr, nodes in attribute_to_nodes.items():
            ret[attr] = [
                DiscoveredHost(
                    hostname=node.metadata.labels["yelp.com/hostname"],
                    pool=node.metadata.labels["yelp.com/pool"],
                )
                for node in nodes
            ]
        return ret


def build_smartstack_location_dict(
    location: str,
    matched_backends_and_tasks: List[
        Tuple[
            Optional[HaproxyBackend],
            Optional[V1Pod],
        ]
    ],
    should_return_individual_backends: bool = False,
) -> MutableMapping[str, Any]:
    running_backends_count = 0
    backends = []
    for backend, task in matched_backends_and_tasks:
        if backend is None:
            continue
        if backend_is_up(backend):
            running_backends_count += 1
        if should_return_individual_backends:
            backends.append(build_smartstack_backend_dict(backend, task))

    return {
        "name": location,
        "running_backends_count": running_backends_count,
        "backends": backends,
    }


def build_smartstack_backend_dict(
    smartstack_backend: HaproxyBackend,
    task: Union[V1Pod],
) -> MutableMapping[str, Any]:
    svname = smartstack_backend["svname"]
    if isinstance(task, V1Pod):
        node_hostname = svname.split("_")[0]
        pod_ip = svname.split("_")[1].split(":")[0]
        hostname = f"{node_hostname}:{pod_ip}"
    else:
        hostname = svname.split("_")[0]
    port = svname.split("_")[-1].split(":")[-1]

    smartstack_backend_dict = {
        "hostname": hostname,
        "port": int(port),
        "status": smartstack_backend["status"],
        "check_status": smartstack_backend["check_status"],
        "check_code": smartstack_backend["check_code"],
        "last_change": int(smartstack_backend["lastchg"]),
        "has_associated_task": task is not None,
    }

    check_duration = smartstack_backend["check_duration"]
    if check_duration:
        smartstack_backend_dict["check_duration"] = int(check_duration)

    return smartstack_backend_dict



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import lru_cache

from botocore.credentials import InstanceMetadataFetcher
from botocore.credentials import InstanceMetadataProvider

from paasta_tools.utils import load_system_paasta_config


try:
    from vault_tools.paasta_secret import get_client as get_vault_client
    from vault_tools.paasta_secret import get_vault_url
    from vault_tools.paasta_secret import get_vault_ca
    from okta_auth import get_and_cache_jwt_default
except ImportError:

    def get_vault_client(url: str, capath: str) -> None:
        pass

    def get_vault_url(ecosystem: str) -> str:
        return ""

    def get_vault_ca(ecosystem: str) -> str:
        return ""

    def get_and_cache_jwt_default(client_id: str) -> str:
        return ""


def get_current_ecosystem() -> str:
    """Get current ecosystem from host configs, defaults to dev if no config is found"""
    try:
        with open("/nail/etc/ecosystem") as f:
            return f.read().strip()
    except IOError:
        pass
    return "devc"


@lru_cache(maxsize=1)
def get_service_auth_token() -> str:
    """Uses instance profile to authenticate with Vault and generate token for service authentication"""
    ecosystem = get_current_ecosystem()
    vault_client = get_vault_client(get_vault_url(ecosystem), get_vault_ca(ecosystem))
    vault_role = load_system_paasta_config().get_service_auth_vault_role()
    metadata_provider = InstanceMetadataProvider(
        iam_role_fetcher=InstanceMetadataFetcher(),
    )
    instance_credentials = metadata_provider.load().get_frozen_credentials()
    vault_client.auth.aws.iam_login(
        instance_credentials.access_key,
        instance_credentials.secret_key,
        instance_credentials.token,
        mount_point="aws-iam",
        role=vault_role,
        use_token=True,
    )
    response = vault_client.secrets.identity.generate_signed_id_token(name=vault_role)
    return response["data"]["token"]


def get_sso_service_auth_token() -> str:
    """Generate an authentication token for the calling user from the Single Sign On provider"""
    client_id = load_system_paasta_config().get_service_auth_sso_oidc_client_id()
    return get_and_cache_jwt_default(client_id)



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# PYTHON_ARGCOMPLETE_OK
"""A command line tool for viewing information from the PaaSTA stack."""
import argparse
import logging
import os
import pkgutil
import subprocess
import sys
import warnings
from typing import Any
from typing import List
from typing import Tuple

import argcomplete

import paasta_tools
from paasta_tools.cli import cmds


def load_method(module_name, method_name):
    """Return a function given a module and method name.

    :param module_name: a string
    :param method_name: a string
    :return: a function
    """
    module = __import__(module_name, fromlist=[method_name])
    method = getattr(module, method_name)
    return method


def modules_in_pkg(pkg):
    """Return the list of modules in a python package (a module with a
    __init__.py file.)

    :return: a list of strings such as `['list', 'check']` that correspond to
             the module names in the package.
    """
    for _, module_name, _ in pkgutil.walk_packages(pkg.__path__):
        yield module_name


class PrintsHelpOnErrorArgumentParser(argparse.ArgumentParser):
    """Overriding the error method allows us to print the whole help page,
    otherwise the python arg parser prints a not-so-useful usage message that
    is way too terse"""

    def error(self, message):
        print(f"Argument parse error: {message}\n")
        self.print_help()
        sys.exit(1)


def list_external_commands():
    p = subprocess.check_output(["/bin/bash", "-p", "-c", "compgen -A command paasta-"])
    lines = p.decode("utf-8").strip().split("\n")
    return {l.replace("paasta-", "", 1) for l in lines}


def calling_external_command():
    if len(sys.argv) > 1:
        return sys.argv[1] in list_external_commands()
    else:
        return False


def exec_subcommand(argv):
    command = sys.argv[1]
    os.execlp(f"paasta-{command}", *argv[1:])


def add_subparser(command, subparsers):
    """Given a command name, paasta_cmd, execute the add_subparser method
    implemented in paasta_cmd.py.

    Each paasta client command must implement a method called add_subparser.
    This allows the client to dynamically add subparsers to its subparser, which
    provides the benefits of argcomplete/argparse but gets it done in a modular
    fashion.

    :param command: a simple string - e.g. 'list'
    :param subparsers: an ArgumentParser object"""
    module_name = "paasta_tools.cli.cmds.%s" % command
    add_subparser_fn = load_method(module_name, "add_subparser")
    add_subparser_fn(subparsers)


PAASTA_SUBCOMMANDS = {
    "autoscale": "autoscale",
    "check": "check",
    "cook-image": "cook_image",
    "get-docker-image": "get_docker_image",
    "get-image-version": "get_image_version",
    "get-latest-deployment": "get_latest_deployment",
    "info": "info",
    "itest": "itest",
    "list-clusters": "list_clusters",
    "list-deploy-queue": "list_deploy_queue",
    "list-namespaces": "list_namespaces",
    "list": "list",
    "local-run": "local_run",
    "logs": "logs",
    "mark-for-deployment": "mark_for_deployment",
    "mesh-status": "mesh_status",
    "pause_service_autoscaler": "pause_service_autoscaler",
    "push-to-registry": "push_to_registry",
    "remote-run": "remote_run",
    "rollback": "rollback",
    "secret": "secret",
    "security-check": "security_check",
    "spark-run": "spark_run",
    "start": "start_stop_restart",
    "stop": "start_stop_restart",
    "restart": "start_stop_restart",
    "status": "status",
    "validate": "validate",
    "wait-for-deployment": "wait_for_deployment",
}


def get_argparser(commands=None):
    """Create and return argument parser for a set of subcommands.

    :param commands: Union[None, List[str]] If `commands` argument is `None`,
    add full parsers for all subcommands, if `commands` is empty list -
    add thin parsers for all subcommands, otherwise - add full parsers for
    subcommands in the argument.
    """

    parser = PrintsHelpOnErrorArgumentParser(
        description=(
            "The PaaSTA command line tool. The 'paasta' command is the entry point "
            "to multiple subcommands, see below.\n\n"
            "You can see more help for individual commands by appending them with '--help', "
            "for example, 'paasta status --help' or see the man page with 'man paasta status'."
        ),
        epilog=(
            "The 'paasta' command line tool is designed to be used by humans, and therefore has "
            "command line completion for almost all options and uses pretty formatting when "
            "possible."
        ),
        # Suppressing usage prevents it from being printed twice upon print_help
        usage=argparse.SUPPRESS,
    )

    # http://stackoverflow.com/a/8521644/812183
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"paasta-tools {paasta_tools.__version__}",
    )

    subparsers = parser.add_subparsers(dest="command", metavar="")
    subparsers.required = True

    # Adding a separate help subparser allows us to respond to "help" without --help
    help_parser = subparsers.add_parser(
        "help", help=f"run `paasta <subcommand> -h` for help"
    )
    help_parser.set_defaults(command=None)

    # Build a list of subcommands to add them in alphabetical order later
    command_choices: List[Tuple[str, Any]] = []
    if commands is None:
        for command in sorted(modules_in_pkg(cmds)):
            command_choices.append(
                (command, (add_subparser, [command, subparsers], {}))
            )
    elif commands:
        for command in commands:
            if command not in PAASTA_SUBCOMMANDS:
                # could be external subcommand
                continue
            command_choices.append(
                (
                    command,
                    (add_subparser, [PAASTA_SUBCOMMANDS[command], subparsers], {}),
                )
            )
    else:
        for command in PAASTA_SUBCOMMANDS.keys():
            command_choices.append(
                (
                    command,
                    (subparsers.add_parser, [command], dict(help="", add_help=False)),
                )
            )

    for command in list_external_commands():
        command_choices.append(
            (command, (subparsers.add_parser, [command], dict(help="")))
        )

    for (_, (fn, args, kwds)) in sorted(command_choices, key=lambda e: e[0]):
        fn(*args, **kwds)

    return parser


def parse_args(argv):
    """Initialize autocompletion and configure the argument parser.

    :return: an argparse.Namespace object mapping parameter names to the inputs
             from sys.argv
    """
    parser = get_argparser(commands=[])
    argcomplete.autocomplete(parser)

    args, _ = parser.parse_known_args(argv)
    if args.command:
        parser = get_argparser(commands=[args.command])

    argcomplete.autocomplete(parser)
    return parser.parse_args(argv), parser


def main(argv=None):
    """Perform a paasta call. Read args from sys.argv and pass parsed args onto
    appropriate command in paasta_cli/cmds directory.

    Ensure we kill any child pids before we quit
    """
    logging.basicConfig()
    warnings.filterwarnings("ignore", category=DeprecationWarning)

    # if we are an external command, we need to exec out early.
    # The reason we exec out early is so we don't bother trying to parse
    # "foreign" arguments, which would cause a stack trace.
    if calling_external_command():
        exec_subcommand(sys.argv)

    try:
        args, parser = parse_args(argv)
        if args.command is None:
            parser.print_help()
            return_code = 0
        else:
            return_code = args.command(args)
    except KeyboardInterrupt:
        return_code = 130
    sys.exit(return_code)


if __name__ == "__main__":
    main()



import sys

import yaml

# try and catch both /opt/venvs/paasta-tools and ~/pg/paasta/.tox/py38-linux as if we're being run as an application,
# we likely want to fail on a potential slowdown rather than experience a performance regression
if "paasta" in sys.prefix:
    from yaml import CSafeLoader as Loader
    from yaml import CSafeDumper as Dumper
# but for the vanishingly few instances where folks add us as as dependency, we don't enforce that they use the
# C-accelerated Loader/Dumper
else:
    try:
        from yaml import CSafeLoader as Loader
        from yaml import CSafeDumper as Dumper
    except ImportError:  # pragma: no cover
        from yaml import SafeLoader as Loader  # type: ignore
        from yaml import SafeDumper as Dumper  # type: ignore


def dump(*args, **kwargs):
    kwargs["Dumper"] = Dumper
    return yaml.dump(*args, **kwargs)


def dump_all(*args, **kwargs):
    kwargs["Dumper"] = Dumper
    return yaml.dump_all(*args, **kwargs)


def load(*args, **kwargs):
    kwargs["Loader"] = Loader
    return yaml.load(*args, **kwargs)


def load_all(*args, **kwargs):
    kwargs["Loader"] = Loader
    return yaml.load_all(*args, **kwargs)


safe_dump = dump
safe_dump_all = dump_all
safe_load = load
safe_load_all = load_all



# coding: utf-8

"""
    Paasta API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.2.0
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from paasta_tools.paastaapi.api_client import ApiClient, Endpoint
from paasta_tools.paastaapi.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    date,
    datetime,
    file_type,
    none_type,
    validate_and_convert_types
)
from paasta_tools.paastaapi.model.remote_run_outcome import RemoteRunOutcome
from paasta_tools.paastaapi.model.remote_run_start import RemoteRunStart
from paasta_tools.paastaapi.model.remote_run_stop import RemoteRunStop
from paasta_tools.paastaapi.model.remote_run_token import RemoteRunToken


class RemoteRunApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        def __remote_run_poll(
            self,
            service,
            instance,
            job_name,
            user,
            **kwargs
        ):
            """Check if remote run pod is ready  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.remote_run_poll(service, instance, job_name, user, async_req=True)
            >>> result = thread.get()

            Args:
                service (str): Service name
                instance (str): Instance name
                job_name (str): Job name
                user (str): User requesting job

            Keyword Args:
                toolbox (bool): Whether this is a toolbox job. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                RemoteRunOutcome
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['service'] = \
                service
            kwargs['instance'] = \
                instance
            kwargs['job_name'] = \
                job_name
            kwargs['user'] = \
                user
            return self.call_with_http_info(**kwargs)

        self.remote_run_poll = Endpoint(
            settings={
                'response_type': (RemoteRunOutcome,),
                'auth': [],
                'endpoint_path': '/remote_run/{service}/{instance}/poll',
                'operation_id': 'remote_run_poll',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'service',
                    'instance',
                    'job_name',
                    'user',
                    'toolbox',
                ],
                'required': [
                    'service',
                    'instance',
                    'job_name',
                    'user',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'service':
                        (str,),
                    'instance':
                        (str,),
                    'job_name':
                        (str,),
                    'user':
                        (str,),
                    'toolbox':
                        (bool,),
                },
                'attribute_map': {
                    'service': 'service',
                    'instance': 'instance',
                    'job_name': 'job_name',
                    'user': 'user',
                    'toolbox': 'toolbox',
                },
                'location_map': {
                    'service': 'path',
                    'instance': 'path',
                    'job_name': 'query',
                    'user': 'query',
                    'toolbox': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__remote_run_poll
        )

        def __remote_run_start(
            self,
            service,
            instance,
            remote_run_start,
            **kwargs
        ):
            """Launch a remote-run pod  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.remote_run_start(service, instance, remote_run_start, async_req=True)
            >>> result = thread.get()

            Args:
                service (str): Service name
                instance (str): Instance name
                remote_run_start (RemoteRunStart):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                RemoteRunOutcome
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['service'] = \
                service
            kwargs['instance'] = \
                instance
            kwargs['remote_run_start'] = \
                remote_run_start
            return self.call_with_http_info(**kwargs)

        self.remote_run_start = Endpoint(
            settings={
                'response_type': (RemoteRunOutcome,),
                'auth': [],
                'endpoint_path': '/remote_run/{service}/{instance}/start',
                'operation_id': 'remote_run_start',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'service',
                    'instance',
                    'remote_run_start',
                ],
                'required': [
                    'service',
                    'instance',
                    'remote_run_start',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'service':
                        (str,),
                    'instance':
                        (str,),
                    'remote_run_start':
                        (RemoteRunStart,),
                },
                'attribute_map': {
                    'service': 'service',
                    'instance': 'instance',
                },
                'location_map': {
                    'service': 'path',
                    'instance': 'path',
                    'remote_run_start': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__remote_run_start
        )

        def __remote_run_stop(
            self,
            service,
            instance,
            remote_run_stop,
            **kwargs
        ):
            """Stop a remote-run pod  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.remote_run_stop(service, instance, remote_run_stop, async_req=True)
            >>> result = thread.get()

            Args:
                service (str): Service name
                instance (str): Instance name
                remote_run_stop (RemoteRunStop):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                RemoteRunOutcome
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['service'] = \
                service
            kwargs['instance'] = \
                instance
            kwargs['remote_run_stop'] = \
                remote_run_stop
            return self.call_with_http_info(**kwargs)

        self.remote_run_stop = Endpoint(
            settings={
                'response_type': (RemoteRunOutcome,),
                'auth': [],
                'endpoint_path': '/remote_run/{service}/{instance}/stop',
                'operation_id': 'remote_run_stop',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'service',
                    'instance',
                    'remote_run_stop',
                ],
                'required': [
                    'service',
                    'instance',
                    'remote_run_stop',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'service':
                        (str,),
                    'instance':
                        (str,),
                    'remote_run_stop':
                        (RemoteRunStop,),
                },
                'attribute_map': {
                    'service': 'service',
                    'instance': 'instance',
                },
                'location_map': {
                    'service': 'path',
                    'instance': 'path',
                    'remote_run_stop': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client,
            callable=__remote_run_stop
        )

        def __remote_run_token(
            self,
            service,
            instance,
            user,
            **kwargs
        ):
            """Get a short lived token for exec into remote-run pod  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.remote_run_token(service, instance, user, async_req=True)
            >>> result = thread.get()

            Args:
                service (str): Service name
                instance (str): Instance name
                user (str): User name

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                RemoteRunToken
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['service'] = \
                service
            kwargs['instance'] = \
                instance
            kwargs['user'] = \
                user
            return self.call_with_http_info(**kwargs)

        self.remote_run_token = Endpoint(
            settings={
                'response_type': (RemoteRunToken,),
                'auth': [],
                'endpoint_path': '/remote_run/{service}/{instance}/token',
                'operation_id': 'remote_run_token',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'service',
                    'instance',
                    'user',
                ],
                'required': [
                    'service',
                    'instance',
                    'user',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'service':
                        (str,),
                    'instance':
                        (str,),
                    'user':
                        (str,),
                },
                'attribute_map': {
                    'service': 'service',
                    'instance': 'instance',
                    'user': 'user',
                },
                'location_map': {
                    'service': 'path',
                    'instance': 'path',
                    'user': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client,
            callable=__remote_run_token
        )
